/****************************************************************************
** libmatroska : parse Matroska files, see http://www.matroska.org/
**
** <file/class description>
**
** Copyright (C) 2002-2003 Steve Lhomme.  All rights reserved.
**
** This file may be distributed under the terms of the Q Public License
** as defined by Trolltech AS of Norway and appearing in the file
** LICENSE.QPL included in the packaging of this file.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** Licensees holding an other license may use this file in accordance with 
** the Agreement provided with the Software.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.matroska.org/license/qpl/ for QPL licensing information.
** See http://www.matroska.org/license/gpl/ for GPL licensing information.
**
** Contact license@matroska.org if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/
/*!
	\file
	\version \$Id: mkx_opin.cpp,v 1.40 2003/05/20 12:11:22 toff Exp $
	\author Steve Lhomme     <robux4 @ users.sf.net>
	\author Christophe Paris <toffparis @ users.sf.net>
*/


#include <streams.h>
#include <initguid.h> // for Vorbis UIDs

#include "mkx_opin.h"
#include "codecs.h"
#include "OggDS.h"
#include "KaxBlockData.h"


void TrackInfos2WaveFormatEx(KaxTrackInfoStruct & TrackInfos, WAVEFORMATEX *pwfx);
void BuildAACDecoderSpecificData(const char* CodecID, int SampleRate,
								 int Channels, char* CodeData);								 

// ***********************************************************
// 
//  The MkxOutPin methods
//

MkxOutPin::MkxOutPin(TCHAR *pObjectName, HRESULT *phr, CSource *pms, LPCWSTR pName, KaxTrackInfoStruct & TrackInfos)
:CSourceStream(pObjectName, phr, pms, pName)
 ,m_TrackInfos(TrackInfos)
 ,m_bSendHeader(false)
 ,mBlockQueue(
	40, // aMaxWriteThreshold
	4, // aMaxReadThreshold
	38, // aMinWriteThreshold
	2) // aMinReadThreshold
 ,mCachedBlockG(NULL)
 ,m_rtStart(0)
{
	unsigned int i;

	switch(TrackInfos.Type)
	{
	case track_video:
		{		
			BITMAPINFOHEADER	*pbmi = (BITMAPINFOHEADER*)TrackInfos.CodecPrivate;
			int sizeOfVH              = sizeof(VIDEOINFOHEADER) + 
				TrackInfos.CodecPrivateLen - sizeof(BITMAPINFOHEADER);
			VIDEOINFOHEADER		*pvih = (VIDEOINFOHEADER*)malloc(sizeOfVH);
			
			m_mt.InitMediaType();
			m_mt.SetType(&MEDIATYPE_Video);
			m_mt.SetFormatType(&FORMAT_VideoInfo);
			memset(pvih, 0, sizeOfVH);
			memcpy(&pvih->bmiHeader, pbmi, TrackInfos.CodecPrivateLen);
			pvih->AvgTimePerFrame = 400000;
			m_mt.SetFormat((BYTE*)pvih, sizeOfVH);
			free(pvih);
			
			FOURCCMap	fcc(pbmi->biCompression);
			m_mt.SetSubtype((GUID*)&fcc);
			m_mt.SetSampleSize(1);			
		}
		break;

	case track_audio:
	{
		GUID			*pGuid = NULL;
		WAVEFORMATEX	*pwfx;
		
		if(strcmp((const char*)TrackInfos.CodecID, "A_VORBIS") == 0)
		{
			VORBISFORMAT vf;
			vf.nChannels = TrackInfos.pAudioInfo->Channels;
			vf.nSamplesPerSec = TrackInfos.pAudioInfo->SamplingFreq;
			vf.nMaxBitsPerSec = -1;    // CHECKME
			vf.nMinBitsPerSec = -1;    // CHECKME
			vf.nAvgBitsPerSec = -1;    // CHECKME
			vf.fQuality = 0;           // CHECKME
			
			m_mt.InitMediaType();
			m_mt.SetType(&MEDIATYPE_Audio);
			m_mt.SetSubtype(&MEDIASUBTYPE_Vorbis);
			m_mt.SetFormatType(&FORMAT_VorbisFormat);
			m_mt.SetFormat((BYTE*)&vf,sizeof(VORBISFORMAT));
			m_mt.SetSampleSize(1);

			m_bSendHeader = true;			
		}
		else if(strcmp((const char*)TrackInfos.CodecID, "A_MS/ACM") == 0)
		{
			WAVEFORMATEX* pwfx = (WAVEFORMATEX*)TrackInfos.CodecPrivate;
			CreateAudioMediaType(pwfx, &m_mt, TRUE);
		}
		else if(strncmp((const char*)TrackInfos.CodecID, "A_AAC/MPEG2", 11) == 0)
		{
			int sizeOfWfx = sizeof(WAVEFORMATEX) + 2;
			pwfx = (WAVEFORMATEX*)malloc(sizeOfWfx);
			memset(pwfx, 0, sizeOfWfx);

			TrackInfos2WaveFormatEx(TrackInfos, pwfx);
			pwfx->wFormatTag = WAVE_FORMAT_AAC;
			pwfx->cbSize = 2;
			char* CodecData = (char*)pwfx+sizeof(WAVEFORMATEX);			
			BuildAACDecoderSpecificData((const char*)TrackInfos.CodecID,
				pwfx->nSamplesPerSec, pwfx->nChannels, CodecData);

			CreateAudioMediaType(pwfx, &m_mt, TRUE);
			
			free(pwfx);
		}
		else
		{
			WAVEFORMATEX	wfx;

			memset(&wfx, 0, sizeof(WAVEFORMATEX));

			TrackInfos2WaveFormatEx(TrackInfos, &wfx);

			for (i=0; i<ACM_CODEC_LIST_LEN; i++)
			{
				if (strcmp((const char*)TrackInfos.CodecID, ACM_CODEC_LIST[i].name) == 0)
				{
					wfx.wFormatTag = ACM_CODEC_LIST[i].id;
					if (ACM_CODEC_LIST[i].id == 0)
					{
						pGuid = (GUID*)&ACM_CODEC_LIST[i].guid;
					}
					break;
				}
			}

			pwfx = &wfx;
			CreateAudioMediaType(pwfx, &m_mt, TRUE);

			if (pGuid != NULL)
			{
				m_mt.SetSubtype(pGuid);					
			}
		}
	}
		break;

	case track_subtitle:
		m_mt.InitMediaType();
		m_mt.SetType(&MEDIATYPE_Text);
		mBlockQueue.SetMinReadThreshold(1);
		mBlockQueue.SetMaxReadThreshold(1);
		break;
	}
}

HRESULT MkxOutPin::GetMediaType(CMediaType *pMediaType)
{
	if (pMediaType == NULL) return E_INVALIDARG;

	pMediaType->InitMediaType();
	pMediaType->SetType(m_mt.Type());
	pMediaType->SetSubtype(m_mt.Subtype());
	pMediaType->SetFormatType(m_mt.FormatType());
	pMediaType->SetTemporalCompression(m_mt.IsTemporalCompressed());
	if(m_mt.FormatLength() > 0)
	{
		pMediaType->SetFormat(m_mt.Format(), m_mt.FormatLength());
	}

	return S_OK;
}

// called from CBaseOutputPin during connection to ask for
// the count and size of buffers we need.
HRESULT MkxOutPin::DecideBufferSize(IMemAllocator *pAlloc, ALLOCATOR_PROPERTIES *pAllocProps)
{
	ALLOCATOR_PROPERTIES	Actual;
	HRESULT	hr;
	
	if (pAlloc == NULL) return E_POINTER;
	if (pAllocProps == NULL) return E_POINTER;
			
	///< \todo check buffer number
	pAllocProps->cBuffers = 8;
	
	// compute buffer size
	switch(m_TrackInfos.Type)
	{
	case track_video:
		if (pAllocProps->cbBuffer < m_TrackInfos.pVideoInfo->PixelHeight *
			m_TrackInfos.pVideoInfo->PixelWidth * 4)
		{
			pAllocProps->cbBuffer = m_TrackInfos.pVideoInfo->PixelHeight *
				m_TrackInfos.pVideoInfo->PixelWidth * 4;
		}
		break;
	case track_audio:
	case track_subtitle:
		///< \todo check buffer size
		pAllocProps->cbBuffer = 0xFFFF;
		break;
	}
	
	pAllocProps->cbAlign = 4;
	
	// set properties
	hr = pAlloc->SetProperties(pAllocProps, &Actual);
	if (FAILED(hr))
		return hr;
	
	return NO_ERROR;
}

/*!
	\todo ensure the whole Block is deleted when not used anymore
*/
HRESULT MkxOutPin::FillBuffer(IMediaSample *pSamp)
{
	CheckPointer(pSamp,E_POINTER);
	BYTE *pData;
	long cbData;
	REFERENCE_TIME	tBlockDuration, tSmpStart, tSmpEnd;	
	uint64 aDuration;

    // Access the sample's data buffer
    pSamp->GetPointer(&pData);
    cbData = pSamp->GetSize();
	
	// Get a block for this track
	if (mCachedBlockG == NULL) 
	{
		while (1) {
			///> \todo GetFrontBlock may remove the block from the queue ?
			mCachedBlockG = mBlockQueue.GetFrontBlock();
			if(mCachedBlockG == NULL)
			{
				return S_FALSE;
			}
			mCachedBlock = static_cast<KaxBlock *>(mCachedBlockG->FindFirstElt(KaxBlock::ClassInfos, false));
			if (mCachedBlock == NULL) {
				delete mCachedBlockG; // continue the search
				mCachedBlockG = NULL;
			} else {
				mPrevStartTime = mCachedBlock->GlobalTimecode() / 100;
				mBlockIndex = 0;
				break;
			}
		}
	}

	// Set sample timestamps
	if (mCachedBlockG->GetBlockDuration(aDuration)) {
		tBlockDuration = aDuration / 100;
	} else { 
		mBlockQueue.Lock();
		if (mBlockQueue[0] != NULL) {
			tBlockDuration = (mBlockQueue[0]->GlobalTimecode() / 100) - mPrevStartTime;
		} else {
			tBlockDuration = 400000; // + 40ms
		}
		mBlockQueue.Unlock();
	}

	if (mBlockIndex + 1 == mCachedBlock->NumberFrames()) {
		tSmpStart = mPrevStartTime + mBlockIndex * (tBlockDuration / mCachedBlock->NumberFrames());
		tSmpEnd = mPrevStartTime + tBlockDuration;
	} else {
		tBlockDuration /= mCachedBlock->NumberFrames();
		tSmpStart = mPrevStartTime + mBlockIndex * tBlockDuration;
		tSmpEnd = tSmpStart + tBlockDuration;
	}
	tSmpStart -= m_rtStart;
	tSmpEnd -= m_rtStart;
	pSamp->SetTime(&tSmpStart, &tSmpEnd);
	//pSamp->SetMediaTime(NULL,NULL);

	// Fill the pData buffer with the sample data
	DataBuffer & aBuf = mCachedBlock->GetBuffer(mBlockIndex);
	CopyMemory(pData, aBuf.Buffer(), aBuf.Size());
	pSamp->SetActualDataLength(aBuf.Size());	
UINT BufSize = aBuf.Size();

	if(m_TrackInfos.Type == track_subtitle)
	{
		NOTE4("Sub start:%I64d, end:%I64d dur:%I64d, %s", tSmpStart, tSmpEnd, tBlockDuration, pData);
	}

	mBlockIndex++;
	if (mBlockIndex != 1 || (NULL == static_cast<KaxReferenceBlock *>(mCachedBlockG->FindFirstElt(KaxReferenceBlock::ClassInfos, false))))
	{ 
		// Set this to true for key frame
		pSamp->SetSyncPoint(TRUE);
	}

	if (mBlockIndex == mCachedBlock->NumberFrames()) {
		delete mCachedBlockG;
		mCachedBlockG = NULL;
		mPrevStartTime = tSmpEnd;
	}
	
	NOTE5("MkxOutPin::FillBuffer out %c 0x%08x %I64d/%I64d buffer size %d", (m_TrackInfos.Type == track_audio)?'A':'V', this, tSmpStart, tSmpEnd, BufSize);
    return NOERROR;
}

HRESULT MkxOutPin::OnThreadStartPlay(void)
{
	NOTE("MkxOutPin::OnThreadStartPlay");
	if(m_bSendHeader)
	{
		NOTE("Send header");
		SendOneHeaderPerSample(m_TrackInfos.CodecPrivate, m_TrackInfos.CodecPrivateLen);
		// we must send header at each start!!!
		//m_bSendHeader = false;
//		DeliverNewSegment( m_rtStart, m_rtStop, 1.0 );
	}
	if (mCachedBlockG != NULL)
		delete mCachedBlockG;
	mCachedBlockG = NULL;

	return NOERROR;
}

HRESULT MkxOutPin::OnThreadDestroy(void)
{
	NOTE("MkxOutPin::OnThreadDestroy");
	if (mCachedBlockG != NULL)
		delete mCachedBlockG;
	mCachedBlockG = NULL;

	return NOERROR;
}

void MkxOutPin::SendOneHeaderPerSample(binary* CodecPrivateData, int DataLen)
{
	IMediaSample	*pSample;
	BYTE			*pData;
	REFERENCE_TIME	tTime = 0;
	binary *codecPrivate = CodecPrivateData;
	uint32 lastHeaderSize = DataLen - 1;
	uint16 nbHeaders = (*(uint8 *)codecPrivate);
	CGenericList<uint32> headerSizeList(_T("Headers"));
	codecPrivate++;
	
	while(nbHeaders--)
	{
		uint32 currentHeaderSize = 0;
		do{
			currentHeaderSize += *(uint8 *)codecPrivate;
			lastHeaderSize--;
		} while((*codecPrivate++) == 0xFF);
		lastHeaderSize -= currentHeaderSize;
		headerSizeList.AddTail((uint32*)currentHeaderSize);
	}
	headerSizeList.AddTail((uint32*)lastHeaderSize);
	
	while(headerSizeList.GetCount()) 
	{
		uint32 currentHeaderSize = (uint32)headerSizeList.GetHead();
		headerSizeList.RemoveHead();
		
		if (FAILED(GetDeliveryBuffer(&pSample, NULL, NULL, 0)))
			return;
		
		pSample->GetPointer(&pData);
		memcpy(pData, codecPrivate, currentHeaderSize);
		pSample->SetActualDataLength(currentHeaderSize);
		pSample->SetTime(&tTime,&tTime);
		pSample->SetMediaTime(&tTime,&tTime);
		Deliver(pSample);
		pSample->Release();
		
		codecPrivate += currentHeaderSize;
	}
}

void MkxOutPin::SetNewStartTime(REFERENCE_TIME aTime)
{
//	CAutoLock(&);
	m_rtStart = aTime;
}

// ----------------------------------------------------------------------------
// Some helper functions
// ----------------------------------------------------------------------------

void TrackInfos2WaveFormatEx(KaxTrackInfoStruct & TrackInfos, WAVEFORMATEX *pwfx)
{
	if(TrackInfos.pAudioInfo->Channels == 0)
		pwfx->nChannels = 1;
	else
		pwfx->nChannels = TrackInfos.pAudioInfo->Channels;
	
	if(TrackInfos.pAudioInfo->SamplingFreq == 0)
		pwfx->nSamplesPerSec = 8000;
	else
		pwfx->nSamplesPerSec = TrackInfos.pAudioInfo->SamplingFreq;	
	
	if (TrackInfos.pAudioInfo->BitDepth == 0)
		pwfx->wBitsPerSample = 16;
	else
		pwfx->wBitsPerSample = TrackInfos.pAudioInfo->BitDepth;
	
	pwfx->nBlockAlign = (unsigned short)((pwfx->nChannels * pwfx->wBitsPerSample) / 8);
	pwfx->nAvgBytesPerSec = pwfx->nSamplesPerSec * pwfx->nBlockAlign;	
}

void BuildAACDecoderSpecificData(const char* CodecID, int SampleRate,
								 int Channels, char* CodeData)
{
	const char* CodecIDProfileString = (const char*)&CodecID[12];
	char profile, srate_idx;
	
	// Recreate the 'private data' which faad2 uses in its initialization.
	// A_AAC/MPEG2/MAIN
	// 0123456789012345
	if (!strcmp(CodecIDProfileString, "MAIN"))
		profile = 0;
	else if (!strcmp(CodecIDProfileString, "LC"))
		profile = 1;
	else if (!strcmp(CodecIDProfileString, "SSR"))
		profile = 2;
	else
		profile = 3;
	
	if (92017 <= SampleRate)
		srate_idx = 0;
	else if (75132 <= SampleRate)
		srate_idx = 1;
	else if (55426 <= SampleRate)
		srate_idx = 2;
	else if (46009 <= SampleRate)
		srate_idx = 3;
	else if (37566 <= SampleRate)
		srate_idx = 4;
	else if (27713 <= SampleRate)
		srate_idx = 5;
	else if (23004 <= SampleRate)
		srate_idx = 6;
	else if (18783 <= SampleRate)
		srate_idx = 7;
	else if (13856 <= SampleRate)
		srate_idx = 8;
	else if (11502 <= SampleRate)
		srate_idx = 9;
	else if (9391 <= SampleRate)
		srate_idx = 10;
	else
		srate_idx = 11;
				
	CodeData[0] = ((profile + 1) << 3) | ((srate_idx & 0xe) >> 1);
	CodeData[1] = ((srate_idx & 0x1) << 7) | (Channels << 3);
}